#ifndef DSPUTIL_H
#define DSPUTIL_H

#include "common.h"
/*
 * Hmm, stdint.h is intended to provide the intNN_t and uintNNNt types in C99,
 * but not everyone has a C99 compatible C environment, yet.  inttypes.h would
 * be a good substitute, it works on linux, xxxBSD and solaris;  but since the
 * code in libavcodec does not use the {u,}intNN_t types, I'll just comment out
 * stdint.h for now.
#include <stdint.h>
 */

/* dct code */
typedef short DCTELEM;

void jpeg_fdct_ifast (DCTELEM *data);

void j_rev_dct (DCTELEM *data);

void fdct_mmx(DCTELEM *block);

/* pixel operations */
#define MAX_NEG_CROP 384

/* temporary */
extern UINT32 squareTbl[512];

void dsputil_init(void);
void get_pixels(DCTELEM *block, const UINT8 *pixels, int line_size);
void put_pixels_clamped(const DCTELEM *block, UINT8 *pixels, int line_size);
void add_pixels_clamped(const DCTELEM *block, UINT8 *pixels, int line_size);

/* add and put pixel (decoding) */
typedef void (*op_pixels_func)(UINT8 *block, const UINT8 *pixels, int line_size, int h);

extern op_pixels_func put_pixels_tab[4];
extern op_pixels_func avg_pixels_tab[4];
extern op_pixels_func put_no_rnd_pixels_tab[4];
extern op_pixels_func avg_no_rnd_pixels_tab[4];

/* sub pixel (encoding) */
extern void (*sub_pixels_tab[4])(DCTELEM *block, const UINT8 *pixels, int line_size, int h);

#define sub_pixels_2(block, pixels, line_size, dxy) \
   sub_pixels_tab[dxy](block, pixels, line_size, 8)

/* motion estimation */
int pix_abs16x16(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_x2(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_y2(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_xy2(unsigned char *blk1, unsigned char *blk2, int lx, int h);

#ifdef CONFIG_MMX
int pix_abs16x16_mmx(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_sse(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_x2_mmx(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_y2_mmx(unsigned char *blk1, unsigned char *blk2, int lx, int h);
int pix_abs16x16_xy2_mmx(unsigned char *blk1, unsigned char *blk2, int lx, int h);

#ifdef CONFIG_SSE
#define pix_abs16x16 pix_abs16x16_sse
#else
#define pix_abs16x16 pix_abs16x16_mmx
#endif
#define pix_abs16x16_x2 pix_abs16x16_x2_mmx
#define pix_abs16x16_y2 pix_abs16x16_y2_mmx
#define pix_abs16x16_xy2 pix_abs16x16_xy2_mmx

static inline void emms(void)
{
	asm volatile ("emms;");
}

#endif

#endif
