template<class X>
static inline const X abs(const X a) {
  return a<0?-a:a;
}

/**
 * Structure that hides some details of proper BITMAPINFOHEADER 
 * initialization methods.
 */
struct BitmapInfo: public BITMAPINFOHEADER
{
    int colors[3];
    BitmapInfo(){}
    BitmapInfo(int width, int height, int bpp)
    {
	memset(this, 0, sizeof(BitmapInfo));
	biWidth=width;
	biHeight=height;
	biPlanes=1;
	switch(bpp)
	{
	case 8:
	case 15:
	case 16:
	case 24:
	case 32:
	    biSizeImage=abs(width*height)*((bpp+7)/8);
	    setBits(bpp);
	    break;
	default:
            printf("DShow: WARNING! invalid bpp in BitmapInfo constructor!\n");
	    setSpace(bpp,bpp);
	    break;
	}
    }
    BitmapInfo(const BITMAPINFOHEADER& hdr)
    {
//	if(hdr.biSize>sizeof(BITMAPINFOHEADER)+12)
//	    throw FATAL("Unknown format");
	memcpy(this, &hdr, hdr.biSize);
    }
    BitmapInfo(const BitmapInfo& bi)
    {
//	if(bi.biSize>sizeof(BITMAPINFOHEADER)+12)
//	    throw FATAL("Unknown format");
	memcpy(this, &bi, bi.biSize);
    }
    BitmapInfo(const BitmapInfo* bi)
    {
//	if(!bi)throw FATAL("Invalid argument");
//	if(bi->biSize>sizeof(BITMAPINFOHEADER)+12)
//	    throw FATAL("Unknown format");
	memcpy(this, bi, bi->biSize);
    }
//    static int BitCount(int csp);
    void setBitFields16()
    {	
	biSize=sizeof(BITMAPINFOHEADER)+12;
	biCompression=3;//BI_BITFIELDS
	biBitCount=16;
	biSizeImage=abs((int)(2*biWidth*biHeight));
	colors[0]=0xF800;
	colors[1]=0x07E0;
	colors[2]=0x001F;
    }	
    void setBitFields15()
    {	
	biSize=sizeof(BITMAPINFOHEADER)+12;
	biCompression=3;//BI_BITFIELDS
	biBitCount=16;
	biSizeImage=abs((int)(2*biWidth*biHeight));
	colors[0]=0x7C00;
	colors[1]=0x03E0;
	colors[2]=0x001F;
    }	
    void setRGB()
    {
	biSize=sizeof(BITMAPINFOHEADER);
	biCompression=0;//BI_RGB
    }
    void setBits(int bits)
    {
	switch(bits)
	{
	    case 15:
		setBitFields15();
		break;
	    case 16:
		setBitFields16();
		break;
	    default:
		setRGB();
		biBitCount=bits;
		break;
	}		
	biSizeImage=abs(biWidth*biHeight)*((biBitCount+7)/8);
    }
    void setSpace(int csp,int bits)
    {
	biCompression=csp;
	biBitCount=bits;
//	biBitCount=BitCount(csp);
	biSizeImage=abs(biBitCount*biWidth*biHeight)>>3;
    }
    int bpp() const
    {
//	if(biCompression==0)return biBitCount;
	if(biCompression!=3)return biBitCount;
	if(colors[0]==0x7c00)return 15;
	return 16;
    }
    bool operator==(const BitmapInfo& bi) const
    {
	return (biWidth==bi.biWidth) 
		&& (biHeight==bi.biHeight)
		&& (biCompression==bi.biCompression)
		&& (bpp() == bi.bpp());
    }
    bool IsRGB() const {return biCompression==0 || biCompression==3;}
};

